@@ -8,9 +8,9 @@ import numpy as np
 import pandas as pd
 from pandas import DataFrame, Int64Dtype
 import networkx as nx
-from sklearn.cluster import KMeans
-from scipy.sparse.csgraph import connected_components
+# from scipy.sparse.csgraph import connected_components
 import warnings
+import bipartitepandas as bpd
 from bipartitepandas import col_order, update_dict, to_list, logger_init, col_dict_optional_cols, aggregate_transform
 
 class BipartiteBase(DataFrame):
@@ -28,7 +28,7 @@ class BipartiteBase(DataFrame):
         include_id_reference_dict (bool): if True, create dictionary of Pandas dataframes linking original id values to contiguous id values
         **kwargs: keyword arguments for Pandas DataFrame
     '''
-    _metadata = ['col_dict', 'reference_dict', 'id_reference_dict', 'col_dtype_dict', 'columns_req', 'columns_opt', 'columns_contig', 'default_KMeans', 'default_cluster', 'dtype_dict', 'default_clean', 'connected', 'correct_cols', 'no_na', 'no_duplicates', 'i_t_unique'] # Attributes, required for Pandas inheritance
+    _metadata = ['col_dict', 'reference_dict', 'id_reference_dict', 'col_dtype_dict', 'columns_req', 'columns_opt', 'columns_contig', 'default_cluster', 'dtype_dict', 'default_clean', 'connected', 'correct_cols', 'no_na', 'no_duplicates', 'i_t_unique'] # Attributes, required for Pandas inheritance
 
     def __init__(self, *args, columns_req=[], columns_opt=[], columns_contig=[], reference_dict={}, col_dtype_dict={}, col_dict=None, include_id_reference_dict=False, **kwargs):
         # Initialize DataFrame
@@ -39,7 +39,7 @@ class BipartiteBase(DataFrame):
         # self.logger.info('initializing BipartiteBase object')
 
         if len(args) > 0 and isinstance(args[0], BipartiteBase): # Note that isinstance works for subclasses
-            self.set_attributes(args[0])
+            self._set_attributes(args[0])
         else:
             self.columns_req = ['i', 'j', 'y'] + columns_req
             self.columns_opt = ['g', 'm'] + columns_opt
@@ -62,32 +62,13 @@ class BipartiteBase(DataFrame):
             self.col_dict = col_dict_optional_cols(default_col_dict, col_dict, self.columns, optional_cols=[self.reference_dict[col] for col in self.columns_opt])
 
             # Set attributes
-            self.reset_attributes()
-
-        # Define default parameter dictionaries
-        self.default_KMeans = { # These parameters are specifically for the KMeans algorithm
-                                # Read more at https://scikit-learn.org/stable/modules/generated/sklearn.cluster.KMeans.html
-            'n_clusters': 10,
-            'init': 'k-means++',
-            'n_init': 500,
-            'max_iter': 300,
-            'tol': 0.0001,
-            'precompute_distances': 'deprecated',
-            'verbose': 0,
-            'random_state': None,
-            'copy_x': True,
-            'n_jobs': 'deprecated',
-            'algorithm': 'auto'
-        }
+            self._reset_attributes()
 
         self.default_cluster = {
-            'cdf_resolution': 10, # How many values to use to approximate the cdfs
-            'grouping': 'quantile_all', # How to group the cdfs ('quantile_all' to get quantiles from entire set of data, then have firm-level values between 0 and 1; 'quantile_firm_small' to get quantiles at the firm-level and have values be compensations if small data; 'quantile_firm_large' to get quantiles at the firm-level and have values be compensations if large data, note that this is up to 50 times slower than 'quantile_firm_small' and should only be used if the dataset is too large to copy into a dictionary; 'mean' to group firms by average income within the firm)
-            'stayers_movers': None, # If None, clusters on entire dataset; if 'stayers', clusters on only stayers; if 'movers', clusters on only movers
-            't': None, # If None, clusters on entire dataset; if int, gives period in data to consider (only valid for non-collapsed data)
-            'weighted': True, # If True, weight firm clusters by firm size (if a weight column is included, firm weight is computed using this column; otherwise, each observation has weight 1)
-            'dropna': False, # If True, drop observations where firms aren't clustered; if False, keep all observations
-            'user_KMeans': self.default_KMeans # Parameters for KMeans estimation (for more information on what parameters can be used, visit https://scikit-learn.org/stable/modules/generated/sklearn.cluster.KMeans.html)
+            'measure_cdf': False, # If True, approximate firm-level income cdfs
+            'measure_moments': False, # (bool, default=False): if True, approximate firm-level moments
+            'cluster_KMeans': False, # (bool, default=False): if True, cluster using KMeans. Valid only if cluster_quantiles=False.
+            'cluster_quantiles': False # If True, cluster using quantiles. Valid only if cluster_KMeans=False, measure_cdf=False, and measure_moments=True, and measuring only a single moment.
         }
 
         self.dtype_dict = {
@@ -118,7 +99,7 @@ class BipartiteBase(DataFrame):
         '''
         df_copy = DataFrame(self, copy=True)
         bdf_copy = self._constructor(df_copy)
-        bdf_copy.set_attributes(self) # This copies attribute dictionaries, default copy does not
+        bdf_copy._set_attributes(self) # This copies attribute dictionaries, default copy does not
 
         return bdf_copy
 
@@ -174,7 +155,7 @@ class BipartiteBase(DataFrame):
         Returns:
             (int or None): number of unique clusters, None if not clustered
         '''
-        if not self.col_included('g'): # If cluster column not in dataframe
+        if not self._col_included('g'): # If cluster column not in dataframe
             return None
         cid_lst = []
         for g_col in to_list(self.reference_dict['g']):
@@ -201,7 +182,7 @@ class BipartiteBase(DataFrame):
         else:
             warnings.warn('id_reference_dict is empty. Either your id columns are already correct, or you did not specify `include_id_reference_dict=True` when initializing your BipartitePandas object')
 
-    def set_attributes(self, frame, no_dict=False):
+    def _set_attributes(self, frame, no_dict=False):
         '''
         Set class attributes to equal those of another BipartitePandas object.
 
@@ -229,12 +210,12 @@ class BipartiteBase(DataFrame):
         self.no_duplicates = frame.no_duplicates # If True, no duplicate rows in the data
         self.i_t_unique = frame.i_t_unique # If True, each worker has at most one observation per period
 
-    def reset_attributes(self):
+    def _reset_attributes(self):
         '''
         Reset class attributes conditions to be False/None.
         '''
         for contig_col in self.columns_contig.keys():
-            if self.col_included(contig_col):
+            if self._col_included(contig_col):
                 self.columns_contig[contig_col] = False
             else:
                 self.columns_contig[contig_col] = None
@@ -245,10 +226,10 @@ class BipartiteBase(DataFrame):
         self.i_t_unique = None # If True, each worker has at most one observation per period; if None, t column not included (set to False later in method if t column included)
 
         # Verify whether period included
-        if self.col_included('t'):
+        if self._col_included('t'):
             self.i_t_unique = False
 
-    def col_included(self, col):
+    def _col_included(self, col):
         '''
         Check whether a column from the pre-established required/optional lists is included.
 
@@ -265,7 +246,7 @@ class BipartiteBase(DataFrame):
             return True
         return False
 
-    def included_cols(self, flat=False):
+    def _included_cols(self, flat=False):
         '''
         Get all columns included from the pre-established required/optional lists.
         
@@ -317,7 +298,7 @@ class BipartiteBase(DataFrame):
                             frame.columns_contig[col] = None
                             if frame.id_reference_dict: # If id_reference_dict has been initialized
                                 frame.id_reference_dict[col] = pd.DataFrame()
-                    elif col not in frame.included_cols() and col not in frame.included_cols(flat=True): # If column is not pre-established
+                    elif col not in frame._included_cols() and col not in frame._included_cols(flat=True): # If column is not pre-established
                         DataFrame.drop(frame, col, axis=1, inplace=True)
                     else:
                         warnings.warn("{} is either (a) a required column and cannot be dropped or (b) a subcolumn that can be dropped, but only by specifying the general column name (e.g. use 'g' instead of 'g1' or 'g2')".format(col))
@@ -325,7 +306,7 @@ class BipartiteBase(DataFrame):
                     warnings.warn('{} is not in data columns'.format(col))
         elif axis == 0:
             DataFrame.drop(frame, indices, axis=0, inplace=True)
-            frame.reset_attributes()
+            frame._reset_attributes()
             frame.clean_data()
 
         return frame
@@ -360,7 +341,7 @@ class BipartiteBase(DataFrame):
                             frame.columns_contig[col_cur] = None
                             if frame.id_reference_dict: # If id_reference_dict has been initialized
                                 frame.id_reference_dict[col_cur] = pd.DataFrame()
-                elif col_cur not in frame.included_cols() and col_cur not in frame.included_cols(flat=True): # If column is not pre-established
+                elif col_cur not in frame._included_cols() and col_cur not in frame._included_cols(flat=True): # If column is not pre-established
                         DataFrame.rename(frame, {col_cur: col_new}, axis=1, inplace=True)
                 else:
                     warnings.warn("{} is either (a) a required column and cannot be renamed or (b) a subcolumn that can be renamed, but only by specifying the general column name (e.g. use 'g' instead of 'g1' or 'g2')".format(col_cur))
@@ -383,10 +364,10 @@ class BipartiteBase(DataFrame):
         frame = DataFrame.merge(self, *args, **kwargs)
         frame = self._constructor(frame) # Use correct constructor
         if kwargs['how'] == 'left': # Non-left merge could cause issues with data, by default resets attributes
-            frame.set_attributes(self)
+            frame._set_attributes(self)
         return frame
 
-    def contiguous_ids(self, id_col):
+    def _contiguous_ids(self, id_col):
         '''
         Make column of ids contiguous.
 
@@ -439,7 +420,7 @@ class BipartiteBase(DataFrame):
 
         return frame
 
-    def update_cols(self, inplace=True):
+    def _update_cols(self, inplace=True):
         '''
         Rename columns and keep only relevant columns.
 
@@ -495,13 +476,13 @@ class BipartiteBase(DataFrame):
         clean_params = update_dict(frame.default_clean, user_clean)
 
         # First, correct columns
-        # Note this must be done before data_validity(), otherwise certain checks are not guaranteed to work
+        # Note this must be done before _data_validity(), otherwise certain checks are not guaranteed to work
         frame.logger.info('correcting columns')
-        frame.update_cols()
+        frame._update_cols()
 
         frame.logger.info('checking quality of data')
-        # Make sure data is valid - computes correct_cols, no_na, no_duplicates, connected, and contiguous, along with other checks (note that column names are corrected in data_validity() if all columns are in the data)
-        frame = BipartiteBase.data_validity(frame) # Shared data_validity
+        # Make sure data is valid - computes correct_cols, no_na, no_duplicates, connected, and contiguous, along with other checks (note that column names are corrected in _data_validity() if all columns are in the data)
+        frame = BipartiteBase._data_validity(frame) # Shared _data_validity
 
         # Next, drop NaN observations
         if not frame.no_na:
@@ -522,23 +503,23 @@ class BipartiteBase(DataFrame):
         # Next, make sure i-t (worker-year) observations are unique
         if frame.i_t_unique is not None and not frame.i_t_unique:
             frame.logger.info('keeping highest paying job for i-t (worker-year) duplicates')
-            frame = frame.drop_i_t_duplicates(how=clean_params['i_t_how'])
+            frame = frame._drop_i_t_duplicates(how=clean_params['i_t_how'])
 
         # Next, find largest set of firms connected by movers
         if not frame.connected:
             # Generate largest connected set
             frame.logger.info('generating largest connected set')
-            frame = frame.conset()
+            frame = frame._conset()
 
         # Next, check contiguous ids
         for contig_col, is_contig in frame.columns_contig.items():
             if is_contig is not None and not is_contig:
                 frame.logger.info('making {} ids contiguous'.format(contig_col))
-                frame = frame.contiguous_ids(contig_col)
+                frame = frame._contiguous_ids(contig_col)
 
         # Using contiguous fids, get NetworkX Graph of largest connected set (note that this must be done even if firms already connected and contiguous)
         # frame.logger.info('generating NetworkX Graph of largest connected set')
-        # _, frame.G = frame.conset(return_G=True) # FIXME currently not used
+        # _, frame.G = frame._conset(return_G=True) # FIXME currently not used
 
         # Sort columns
         frame.logger.info('sorting columns')
@@ -549,7 +530,7 @@ class BipartiteBase(DataFrame):
 
         return frame
 
-    def data_validity(self):
+    def _data_validity(self):
         '''
         Checks that data is formatted correctly and updates relevant attributes.
 
@@ -561,7 +542,7 @@ class BipartiteBase(DataFrame):
         success = True
 
         frame.logger.info('--- checking columns ---')
-        all_cols = frame.included_cols()
+        all_cols = frame._included_cols()
         cols = True
         frame.logger.info('--- checking column datatypes ---')
         col_dtypes = True
@@ -578,7 +559,7 @@ class BipartiteBase(DataFrame):
                         frame.logger.info('{} has wrong dtype, should be {} but is {}'.format(frame.col_dict[subcol], frame.col_dtype_dict[col], col_type))
                         if col in frame.columns_contig.keys(): # If column contiguous
                             frame.logger.info('{} has wrong dtype, converting to contiguous integers'.format(frame.col_dict[subcol]))
-                            frame = frame.contiguous_ids(col)
+                            frame = frame._contiguous_ids(col)
                         else:
                             frame.logger.info('{} has wrong dtype. Please check that this is the correct column (it is supposed to give {}), and if it is, cast it to a valid datatype (these include: {})'.format(frame.col_dict[subcol], subcol, valid_types))
                             col_dtypes = False
@@ -603,7 +584,7 @@ class BipartiteBase(DataFrame):
             success = False
 
         frame.logger.info('--- checking non-pre-established columns ---')
-        all_cols = frame.included_cols(flat=True)
+        all_cols = frame._included_cols(flat=True)
         for col in frame.columns:
             if col not in all_cols:
                 cols = False
@@ -635,7 +616,7 @@ class BipartiteBase(DataFrame):
         else:
             frame.no_duplicates = True
 
-        if frame.col_included('t'):
+        if frame._col_included('t'):
             frame.logger.info('--- checking i-t (worker-year) observations ---')
             max_obs = 1
             for t_col in to_list(frame.reference_dict['t']):
@@ -675,7 +656,7 @@ class BipartiteBase(DataFrame):
 
         # Check contiguous columns
         for contig_col, is_contig in frame.columns_contig.items():
-            if frame.col_included(contig_col):
+            if frame._col_included(contig_col):
                 frame.logger.info('--- checking contiguous {} ids ---'.format(contig_col))
                 id_max = - np.inf
                 ids_unique = []
@@ -697,7 +678,7 @@ class BipartiteBase(DataFrame):
 
         return frame
 
-    def drop_i_t_duplicates(self, how='max'):
+    def _drop_i_t_duplicates(self, how='max'):
         '''
         Keep only the highest paying job for i-t (worker-year) duplicates.
 
@@ -709,7 +690,7 @@ class BipartiteBase(DataFrame):
         '''
         frame = self.copy()
 
-        if frame.col_included('t'):
+        if frame._col_included('t'):
             try: # If BipartiteEventStudy or BipartiteEventStudyCollapsed
                 frame = frame.unstack_es().drop('m').drop_duplicates()
                 convert_1 = True
@@ -745,7 +726,7 @@ class BipartiteBase(DataFrame):
 
         return frame
 
-    def conset(self, return_G=False):
+    def _conset(self, return_G=False):
         '''
         Update data to include only the largest connected set of movers, and if firm ids are contiguous, also return the NetworkX Graph.
 
@@ -821,7 +802,7 @@ class BipartiteBase(DataFrame):
         else:
             frame = self.copy()
 
-        if not frame.col_included('m'):
+        if not frame._col_included('m'):
             if len(to_list(frame.reference_dict['j'])) == 1:
                 frame['m'] = (aggregate_transform(frame, col_groupby='i', col_grouped='j', func='n_unique', col_name='m') > 1).astype(int)
             elif len(to_list(frame.reference_dict['j'])) == 2:
@@ -838,30 +819,28 @@ class BipartiteBase(DataFrame):
 
         return frame
 
-    def approx_cdfs(self, cdf_resolution=10, grouping='quantile_all', stayers_movers=None, t=None, weighted=True):
+    def _prep_cluster_data(self, stayers_movers=None, t=None, weighted=True):
         '''
-        Generate cdfs of compensation for firms.
+        Prepare data for clustering.
 
         Arguments:
-            cdf_resolution (int): how many values to use to approximate the cdfs (when grouping by 'mean', this gives the number of quantiles to compute)
-            grouping (str): how to group the cdfs ('quantile_all' to get quantiles from entire set of data, then have firm-level values between 0 and 1; 'quantile_firm_small' to get quantiles at the firm-level and have values be compensations if small data; 'quantile_firm_large' to get quantiles at the firm-level and have values be compensations if large data, note that this is up to 50 times slower than 'quantile_firm_small' and should only be used if the dataset is too large to copy into a dictionary; 'mean' to group firms by average income within the firm)
-            stayers_movers (str or None): if None, uses entire dataset; if 'stayers', uses only stayers; if 'movers', uses only movers
-            t (int or None): if None, clusters on entire dataset; if int, gives period in data to consider (only valid for non-collapsed data)
-            weighted (bool): if True, compute firm weights
-
+            stayers_movers (str or None, default=None): if None, clusters on entire dataset; if 'stayers', clusters on only stayers; if 'movers', clusters on only movers
+            t (int or None, default=None): if None, clusters on entire dataset; if int, gives period in data to consider (only valid for non-collapsed data)
+            weighted (bool, default=True): if True, weight firm clusters by firm size (if a weight column is included, firm weight is computed using this column; otherwise, each observation has weight 1)
         Returns:
-            cdf_df (NumPy Array): NumPy array of firm cdfs
+            data (Pandas DataFrame): data prepared for clustering
             weights (NumPy Array or None): if weighted=True, gives NumPy array of firm weights for clustering; otherwise, is None
             jids (NumPy Array): firm ids of firms in subset of data used to cluster
         '''
+        # Prepare data
         if stayers_movers is not None: # Note this condition is split into two sections, one prior to stacking event study data and one post
             # Generate m column (the function checks if it already exists)
             self.gen_m()
 
         # Stack data if event study (need all data in 1 column)
-        if len(to_list(self.reference_dict['j'])) == 2:
+        try:
             frame = self.get_long(return_df=True) # Returns Pandas dataframe, not BipartiteLong(Collapsed)
-        else:
+        except AttributeError:
             frame = self
 
         if stayers_movers is not None:
@@ -879,158 +858,69 @@ class BipartiteBase(DataFrame):
             else:
                 warnings.warn('Cannot use data from a particular period on collapsed data, proceeding to cluster on all data')
 
-        # Create return variables
-        jids = sorted(data['j'].unique()) # Must sort
-        n_firms = len(jids) # Can't use self.n_firms() since data could be a subset of self.data
-        if grouping == 'mean':
-            cdfs = np.zeros(n_firms)
+        # Create weights
+        if weighted:
+            if self._col_included('w'):
+                data['row_weights'] = data['w']
+                weights = data.groupby('j')['w'].sum().to_numpy()
+            else:
+                data['row_weights'] = 1
+                weights = data.groupby('j').size().to_numpy()
         else:
-            cdfs = np.zeros([n_firms, cdf_resolution])
-
-        # Create quantiles of interest
-        quantiles = np.linspace(1 / cdf_resolution, 1, cdf_resolution)
-
-        if grouping == 'mean':
-            # Group by mean income
+            data['row_weights'] = 1
             weights = None
-            if weighted and self.col_included('w'):
-                data['weighted_y'] = data['y'] * data['w']
-                data['weighted_w'] = data['w']
-            else:
-                data['weighted_y'] = data['y']
-                data['weighted_w'] = 1
-            firm_mean = data.groupby('j')['weighted_y'].sum() / data.groupby('j')['weighted_w'].sum()
-            # Drop columns
-            data.drop(['weighted_y', 'weighted_w'], axis=1, inplace=True)
-            self.drop(['weighted_y', 'weighted_w']) # Because data is not a copy, must also drop from self
-            # Get quantiles for average firm income
-            quantile_groups = firm_mean.quantile(quantiles)
-            for i, mean_income in enumerate(firm_mean):
-                # Find quantile for each firm
-                quantile_group = 0
-                for quantile in quantile_groups:
-                    if mean_income > quantile:
-                        quantile_group += 1
-                    else:
-                        break
-                cdfs[i] = quantile_group
 
-        else:
-            # Group by income cdfs
-            # Create weights
-            if weighted:
-                if self.col_included('w'):
-                    weights = data.groupby('j')['w'].sum()
-                else:
-                    weights = data.groupby('j').size()
-            else:
-                weights = None
+        # Get unique firm ids
+        jids = sorted(data['j'].unique()) # Must sort
 
-            if grouping == 'quantile_all':
-                # Get quantiles from all data
-                quantile_groups = data['y'].quantile(quantiles)
+        return data, weights, jids
 
-                # Generate firm-level cdfs
-                data.sort_values('j', inplace=True) # Required for aggregate_transform
-                for i, quant in enumerate(quantile_groups):
-                    data['quant'] = (data['y'] <= quant).astype(int)
-                    cdfs_col = aggregate_transform(data, col_groupby='j', col_grouped='quant', func='sum', merge=False) # aggregate(data['fid'], firm_quant, func='sum', fill_value=- 1)
-                    cdfs[:, i] = cdfs_col[cdfs_col >= 0]
-                data.drop('quant', axis=1, inplace=True)
-                del cdfs_col
-
-                # Normalize by firm size (convert to cdf)
-                jsize = data.groupby('j').size().to_numpy()
-                cdfs = (cdfs.T / jsize.T).T
-
-            elif grouping in ['quantile_firm_small', 'quantile_firm_large']:
-                # Sort data by compensation (do this once now, so that don't need to do it again later) (also note it is faster to sort then manually compute quantiles than to use built-in quantile functions)
-                data = data.sort_values(['y'])
-
-                if grouping == 'quantile_firm_small':
-                    # Convert pandas dataframe into a dictionary to access data faster
-                    # Source for idea: https://stackoverflow.com/questions/57208997/looking-for-the-fastest-way-to-slice-a-row-in-a-huge-pandas-dataframe
-                    # Source for how to actually format data correctly: https://stackoverflow.com/questions/56064677/pandas-series-to-dict-with-repeated-indices-make-dict-with-list-values
-                    # data_dict = data['y'].groupby(level=0).agg(list).to_dict()
-                    data_dict = data.groupby('j')['y'].agg(list).to_dict()
-                    # data.sort_values(['j', 'y'], inplace=True) # Required for aggregate_transform
-                    # data_dict = pd.Series(aggregate_transform(data, col_groupby='j', col_grouped='y', func='array', merge=False), index=np.unique(data['j'])).to_dict()
-                    # with warnings.catch_warnings():
-                    #     warnings.filterwarnings('ignore', category=np.VisibleDeprecationWarning)
-                    #     data_dict = pd.Series(aggregate(data['j'], data['y'], func='array', fill_value=[]), index=np.unique(data['j'])).to_dict()
-
-                # Generate the cdfs
-                for i, jid in enumerate(jids):
-                    # Get the firm-level compensation data (don't need to sort because already sorted)
-                    if grouping == 'quantile_firm_small':
-                        y = data_dict[jid]
-                    elif grouping == 'quantile_firm_large':
-                        y = data.loc[data['j'] == jid, 'y'].to_numpy()
-                    # Generate the firm-level cdf
-                    # Note: update numpy array element by element
-                    # Source: https://stackoverflow.com/questions/30012362/faster-way-to-convert-list-of-objects-to-numpy-array/30012403
-                    for j in range(cdf_resolution):
-                        index = max(len(y) * (j + 1) // cdf_resolution - 1, 0) # Don't want negative index
-                        # Update cdfs with the firm-level cdf
-                        cdfs[i, j] = y[index]
-
-        return cdfs, weights, jids
-
-    def cluster(self, user_cluster={}):
+    def cluster(self, measures=bpd.measures.cdfs(), grouping=bpd.grouping.kmeans(), stayers_movers=None, t=None, weighted=True, dropna=False):
         '''
         Cluster data and assign a new column giving the cluster for each firm.
 
         Arguments:
-            user_cluster (dict): dictionary of parameters for clustering
-
-                Dictionary parameters:
-
-                    cdf_resolution (int, default=10): how many values to use to approximate the cdfs (when grouping by 'mean', this gives the number of quantiles to compute)
-
-                    grouping (str, default='quantile_all'): how to group the cdfs ('quantile_all' to get quantiles from entire set of data, then have firm-level values between 0 and 1; 'quantile_firm_small' to get quantiles at the firm-level and have values be compensations if small data; 'quantile_firm_large' to get quantiles at the firm-level and have values be compensations if large data, note that this is up to 50 times slower than 'quantile_firm_small' and should only be used if the dataset is too large to copy into a dictionary; 'mean' to group firms by average income within the firm)
-
-                    stayers_movers (str or None, default=None): if None, clusters on entire dataset; if 'stayers', clusters on only stayers; if 'movers', clusters on only movers
-
-                    t (int or None, default=None): if None, clusters on entire dataset; if int, gives period in data to consider (only valid for non-collapsed data)
-
-                    weighted (bool, default=True): if True, weight firm clusters by firm size (if a weight column is included, firm weight is computed using this column; otherwise, each observation has weight 1)
-
-                    dropna (bool, default=False): if True, drop observations where firms aren't clustered; if False, keep all observations
-
-                    user_KMeans (dict): parameters for KMeans estimation (for more information on what parameters can be used, visit https://scikit-learn.org/stable/modules/generated/sklearn.cluster.KMeans.html)
+            measures (function or list of functions): how to compute measures for clustering. Options can be seen in bipartitepandas.measures.
+            grouping (function): how to group firms based on measures. Options can be seen in bipartitepandas.grouping.
+            stayers_movers (str or None, default=None): if None, clusters on entire dataset; if 'stayers', clusters on only stayers; if 'movers', clusters on only movers
+            t (int or None, default=None): if None, clusters on entire dataset; if int, gives period in data to consider (only valid for non-collapsed data)
+            weighted (bool, default=True): if True, weight firm clusters by firm size (if a weight column is included, firm weight is computed using this column; otherwise, each observation has weight 1)
+            dropna (bool, default=False): if True, drop observations where firms aren't clustered; if False, keep all observations
 
         Returns:
             frame (BipartiteBase): BipartiteBase with clusters
         '''
         frame = self.copy()
 
-        # Update dictionary
-        cluster_params = update_dict(frame.default_cluster, user_cluster)
+        # Prepare data for clustering
+        cluster_data, weights, jids = self._prep_cluster_data(stayers_movers=stayers_movers, t=t, weighted=weighted)
 
-        # Unpack dictionary
-        cdf_resolution = cluster_params['cdf_resolution']
-        grouping = cluster_params['grouping']
-        stayers_movers = cluster_params['stayers_movers']
-        t = cluster_params['t']
-        weighted = cluster_params['weighted']
-        user_KMeans = cluster_params['user_KMeans']
+        # Compute measures
+        for i, measure in enumerate(to_list(measures)):
+            if i == 0:
+                computed_measures = measure(cluster_data, jids)
+            else:
+                # For computing both cdfs and moments
+                computed_measures = np.concatenate([computed_measures, measure(cluster_data, jids)], axis=1)
+        frame.logger.info('firm moments computed')
 
-        # Compute cdfs
-        cdfs, weights, jids = frame.approx_cdfs(cdf_resolution=cdf_resolution, grouping=grouping, stayers_movers=stayers_movers, t=t, weighted=weighted)
-        frame.logger.info('firm cdfs computed')
+        # Drop columns (because prepared data is not always a copy, must drop from self)
+        for col in ['row_weights', 'one']:
+            if col in self.columns:
+                self.drop(col)
 
-        if grouping == 'mean':
-            # No KMeans needed
-            clusters = cdfs
-        else:
-            # Compute firm clusters
-            KMeans_params = update_dict(frame.default_KMeans, user_KMeans)
-            frame.logger.info('computing firm clusters')
-            clusters = KMeans(**KMeans_params).fit(cdfs, sample_weight=weights).labels_
-            frame.logger.info('firm clusters computed')
+        # Can't group using quantiles if more than 1 column
+        if (grouping.__name__ == 'compute_quantiles') and (computed_measures.shape[1] > 1):
+            grouping = bpd.measures.kmeans()
+            warnings.warn('Cannot cluster using quantiles if multiple measures computed. Defaulting to KMeans.')
+
+        # Compute firm groups
+        frame.logger.info('computing firm groups')
+        clusters = grouping(computed_measures, weights)
+        frame.logger.info('firm groups computed')
 
         # Drop existing clusters
-        if frame.col_included('g'):
+        if frame._col_included('g'):
             frame.drop('g')
 
         for i, j_col in enumerate(to_list(frame.reference_dict['j'])):
@@ -1052,7 +942,7 @@ class BipartiteBase(DataFrame):
         sorted_cols = sorted(frame.columns, key=col_order)
         frame = frame[sorted_cols]
 
-        if cluster_params['dropna']:
+        if dropna:
             # Drop firms that don't get clustered
             frame = frame.dropna().reset_index(drop=True)
             frame[frame.reference_dict['g']] = frame[frame.reference_dict['g']].astype(int)
